from abc import ABC, abstractmethod

import msgspec

__all__ = (
    "AbstractJSONDecoder",
    "MSGSpecDecoder",
)


class AbstractJSONDecoder(ABC):
    """JSON decoder interface"""

    @staticmethod
    @abstractmethod
    def serialize[T](data: bytes, model: type[T]) -> T:
        """Serializes a raw bytes object into a given data structure model
        :param data: Raw bytes to serialize
        :param model: Model used for serialization
        :return: Serialized instance of a given model
        """


class MSGSpecDecoder(AbstractJSONDecoder):
    """JSON decoder that uses the msgspec library"""

    @staticmethod
    def serialize[T: msgspec.Struct](data: bytes, model: type[T]) -> T:
        """Serializes a raw bytes object into a given data structure model
        :param data: Raw bytes to serialize
        :param model: Model used for serialization
        :return: Serialized instance of a given model
        """
        return msgspec.json.decode(data, type=model)



"""NBT (Named Binary Tag) protocol decoder
This relies on the https://pypi.org/project/NBT/ library.
"""

import json as dum_json
from base64 import b64decode
from io import BytesIO
from typing import Any

import nbt
from msgspec import Struct

__all__ = ("DecodedNBT",)


def _decode_nbt(raw_data) -> Any:
    return nbt.nbt.NBTFile(fileobj=BytesIO(b64decode(raw_data)))[0][0]


class DecodedNBT(Struct):
    """Decoded NBT item data"""

    raw_data: str
    is_pet: bool = False
    skyblock_id: str = ""
    reforge: str | None = None
    dungeon_stars: int = 0
    hot_potato_count: int = 0
    rarity_upgrades: int = 0
    art_of_war_count: int = 0
    gems: list[str] | None = []
    scrolls: list[str] | None = []
    enchantments: list[str] | None = []

    def __post_init__(self) -> None:
        if self.raw_data == "":
            return
        nbt_data = _decode_nbt(self.raw_data)
        ea = nbt_data["tag"]["ExtraAttributes"]

        self.skyblock_id = str(ea["id"]).replace("STARRED_", "")

        if self.skyblock_id == "PET":
            self.is_pet = True
            pet_info = dum_json.loads(ea["petInfo"].value)
            pet_type = pet_info["type"]
            self.skyblock_id = f"{pet_type}_PET"

        self.reforge = None if ea.get("modifier") is None else ea.get("modifier").value

        if ea.get("dungeon_item_level") is not None:
            self.dungeon_stars = ea.get("dungeon_item_level").value
        elif ea.get("upgrade_level") is not None:
            self.dungeon_stars = ea.get("upgrade_level").value
        else:
            self.dungeon_stars = 0

        self.hot_potato_count = 0 if ea.get("hot_potato_count") is None else int(ea.get("hot_potato_count").value)
        self.rarity_upgrades = 0 if ea.get("rarity_upgrades") is None else int(ea.get("rarity_upgrades").value)

        self.enchantments = []
        if ea.get("enchantments") is not None:
            for enchantment_tag in ea.get("enchantments").tags:
                self.enchantments.append(
                    "ENCHANTMENT_" + enchantment_tag.name.upper() + "_" + str(enchantment_tag.value)
                )

        self.art_of_war_count = 0
        if ea.get("art_of_war_count") is not None:
            self.art_of_war_count = int(ea.get("art_of_war_count").value)

        self.gems = []
        if ea.get("gems") is not None:
            self.gems = ea.get("gems").value

        self.scrolls = []
        if ea.get("ability_scroll") is not None:
            for scroll_tag in ea.get("ability_scroll").tags:
                self.scrolls.append(scroll_tag.value)

    def dict(self) -> dict:
        """Creates a dictionary representation of the NBT data.

        This is a representation of the NBT structure as it is parsed by the NBT library.
        Some fields of the returned dictionary may be different from the instance attributes
        of the `DecodedNBT` class.
        """

        def _build_dict(node: Any) -> dict:
            """Recursive helper function"""
            d = {}
            for key in list(node):
                tag = node.get(key)
                if tag.id == 10:  # Compound tag
                    d[key] = _build_dict(tag)
                else:
                    d[key] = tag.value
            return d

        nbt_data = _decode_nbt(self.raw_data)
        return _build_dict(nbt_data)



import pytest

from reskyblock import Client, MockClient


@pytest.fixture
def client() -> MockClient:
    return MockClient()


@pytest.mark.asyncio
async def test_client() -> None:
    _ = MockClient()


@pytest.mark.asyncio
async def test_get_auctions(client: Client) -> None:
    auctions = await client.get_auctions()
    assert auctions.auctions[0].decoded_nbt is not None
    assert auctions.auctions[0].decoded_nbt.skyblock_id == "MOCK_ITEM"


@pytest.mark.asyncio
async def test_get_auctions_ended(client: Client) -> None:
    auctions_ended = await client.get_auctions_ended()
    assert auctions_ended.auctions[0].decoded_nbt is not None
    assert auctions_ended.auctions[0].decoded_nbt.skyblock_id == "MOCK_ITEM"


@pytest.mark.asyncio
async def test_get_bazaar(client: Client) -> None:
    assert (await client.get_bazaar()).products["MOCK_PRODUCT"].quick_status.sell_price == 1.0


@pytest.mark.asyncio
async def test_continuous_update_auctions(client: Client) -> None:
    async for auctions in await client.get_auctions_continuous():
        assert len(auctions.auctions) == 1
    assert True


@pytest.mark.asyncio
async def test_continuous_update_auctions_ended(client: Client) -> None:
    async for ended_auctions in await client.get_auctions_ended_continuous():
        assert len(ended_auctions.auctions) == 1
    assert True


@pytest.mark.asyncio
async def test_continuous_update_bazaar(client: Client) -> None:
    async for bazaar in await client.get_bazaar_continuous():
        assert len(bazaar.products) == 1
    assert True


@pytest.mark.asyncio
async def test_get_all_auctions(client: Client) -> None:
    auctions = await client.get_all_auctions()
    assert len(auctions.auctions) == 1


@pytest.mark.asyncio
async def test_get_all_auctions_continuous(client: Client) -> None:
    n_iter = 0
    async for all_auctions in await client.get_all_auctions_continuous():
        n_iter += 1
        assert len(all_auctions.auctions) == 1
    assert n_iter == 1



import msgspec

__all__ = (
    "Bazaar",
    "Product",
    "QuickStatus",
    "Summary",
)


class Summary(msgspec.Struct, rename="camel"):
    amount: int
    price_per_unit: float
    orders: int


class QuickStatus(msgspec.Struct, rename="camel"):
    product_id: str
    sell_price: float
    sell_volume: int
    sell_moving_week: int
    sell_orders: int
    buy_price: float
    buy_volume: int
    buy_moving_week: int
    buy_orders: int


class Product(msgspec.Struct):
    product_id: str
    sell_summary: list[Summary]
    buy_summary: list[Summary]
    quick_status: QuickStatus


class Bazaar(msgspec.Struct, rename="camel"):
    success: bool
    last_updated: int
    products: dict[str, Product]
    received_at: float = 0



from reskyblock.http.http_adapter import AbstractAsyncHTTPClient, AbstractHTTPClient, HTTPXAsyncClient, HTTPXClient

__all__ = ("HTTPXClient", "AbstractHTTPClient", "HTTPXAsyncClient", "AbstractAsyncHTTPClient")



from reskyblock.models.auctions import AllAuctions, Auction, Auctions
from reskyblock.models.auctions_ended import AuctionsEnded, EndedAuction
from reskyblock.models.bazaar import Bazaar, Product, QuickStatus, Summary

__all__ = (
    "Auction",
    "Auctions",
    "Bazaar",
    "Product",
    "AuctionsEnded",
    "EndedAuction",
    "AllAuctions",
    "QuickStatus",
    "Summary",
)



from reskyblock.client import AbstractClient, Client, MockClient
from reskyblock.models import (
    AllAuctions,
    Auction,
    Auctions,
    AuctionsEnded,
    Bazaar,
    EndedAuction,
    Product,
    QuickStatus,
    Summary,
)
from reskyblock.nbt import DecodedNBT

__all__ = (
    "Client",
    "DecodedNBT",
    "Auction",
    "EndedAuction",
    "Auctions",
    "AuctionsEnded",
    "Bazaar",
    "Product",
    "MockClient",
    "AbstractClient",
    "QuickStatus",
    "Summary",
    "AllAuctions",
)



import pytest

from reskyblock.nbt import DecodedNBT

_BULK_NBTS: list[str] = (
    (__import__("pathlib").Path(__file__).resolve().parents[0] / "data" / "item_bytes.txt").read_text().split("\n")[:-1]
)


def test_item_simple() -> None:
    raw_data = "H4sIAAAAAAAAAE1SzW7aQBAeIGkAqWmPlXqopaa3uLJxwPjQgws0GGG7hL/Yl2rZXWDJro3sNQm8Q54jL9AniPpcVZfk0tNov5/5dkZTB6hBidUBoFSGMiOlxxKcdtIikaU6VCRaVaDWZ4T+4GiVK9XfOtTHdwXn4X1CsyqUPQIXJrathdXEOsXE0q2l7ejIdpCOW21qN9pX5qLZUr6fWbqlmWQ0r0FV0gdZZDR/ia7C6QzxgsJvuh8Y8e3aILcDjvdeS70nY4OH3mZre8lsv+h4LU8ovu+2hnvnP21TonmTR9ZgHSejYiFmxtC64bR/Y2Ix3UUH1wqvo0O88Yxo07OCzXcRzkdGNB8wv7sy/fmMKb4RHchdLGIRKiyY4Pvwetrwu6Nm3O2Z8bV/729cwxeDddgdHQI2cJa3xjc1QR3OCMu3HO1rcDJMM1pV4Bl8fn6y1dQ7tcBcY0IUCZN7Da0QS3KpESTQisJHJVpmqdCen1KRKvz5CQ/RDin4q+piqTpfM041lmhyTY90J2MiTxPNyzlV9KW2ylAicw0pEjWML/BJVUW8JmgZJQWWTBmOuRSRY98PLxI+DTqh74eB5nY6vfE4vImqcBIgQeG9Ehz/oU0QZ7lAiZryXe9BZsiVMmOLQtK8erwZeDt0Z+6viTv0xr4bKH9RKPSiQRzcNBaGbpOWqV/ZiOiOg9u6YZlLSjGy8ZKeQE0yQXOJxFYdwuP55fkfgDK86b4upwLwD93/ZTikAgAA"
    did = DecodedNBT(raw_data)
    assert did
    assert did.skyblock_id == "LAVA_TALISMAN"
    assert did.enchantments == []
    assert did.art_of_war_count == 0
    assert did.rarity_upgrades == 0
    assert not did.is_pet
    assert did.reforge is None
    assert did.dungeon_stars == 0
    assert did.scrolls == []


def test_item_pet() -> None:
    raw_data = "H4sIAAAAAAAAAI1Uz2/iRhQekmwXUKVVVyv1x2lqbQ/V4nRsY4Mj9cAGShwBYVmSAFUPg/3Aztpjao8ToOqfsD1WPfTOsarUQ8/8Kfkfel31mURRpF4qjTR+P773fe+NZ8qElEghKBNCCntkL/AK7wvkyXGcCVkok33J5/ukdBJ48F3I5ylmfSiT8tt3WRie3QhIimTP8chL3TJZXa9Zqss0rho6t1VbZ5pa8zhoNb1qm1VAXD+JF5DIANISKUpYyiyBdEddJE8ueJgB+QtWp2wy8pk3Og3dlWOhPXzLwjPnalFzxMVqeuxYToTxk4bVWdmPck3JL81wbJz6E/Emm0YXrGMMQjgZaG50ft29cpbdpqP1ml3Wbb7RxtHA760bq96ls54Mz5e9trPuDic+7kZv/fqqF0388XAQjIehP2k7bBydr8+aDbM3dPRJnqOPzcnxqT0bsW+xgzJ56gXpIuSrEjnoxAkU0fkp+WS7qbeE63MhAzGnfZDofr7d1JowA5HCEd1u+CuTkc/Q5wgJYRjMQbj3AY0xzP94u7E6PIVkCjwiGmbimvBFSldxllAQEOFI6SxOEDTV2RK32i6E5u3vvxAF98fF8zhcQ7LKSYwUrUOk+QJpHmm9DFIvjujrOE4l+eaOFFc7wXiKQOOVwW5/+4P+B3JfLlfd5YLTfhyH5PAOPQAUAAmXuYapwb7aSVnKhNMIcyvUi7NpCB75HP03PggqgCcUOwsEvUFYkpd+htjtJuw2RrTTumh1yAsc8+2vf1PdrNh1raLVGR31MfHrHSsMgrkvVTcM3HdUxpR7HpV+kNIFyNzOR5XLhdyOQGRf3nFYyNFptVu9ZmMwLpKDHo9gd1Dfd65DimfzA6q32hlPvIAL/AWetfJOGlImwTSTkBbJUyzpiFlM/vlJkasFKEdK+7wxaDqNnlJRuCuDa/TNeJhCRYHlQjnSD01sATuoG6ZhGLpttmoVBe9MgtgHNQj28VLmpR/gLhfe6jwFTzliFSXLAvxQ7NpU03kd1JluW2qVV+uqDTZTTV616vWqZ7AZw2KZCH7MwMkRVeCeVp9OVdPVa2pVY5Zqc3emzpjF8SJzw+DVe/rdXI/zsT6IEHE3fmjp52L+opD9fmuI48sVkZf/R9ABKckgglTyaIGPw/vnH/50CNkjHzV5xOdA9gn5F2U272a4BAAA"
    did = DecodedNBT(raw_data)
    assert did
    assert did.skyblock_id == "GUARDIAN_PET"
    assert did.enchantments == []
    assert did.art_of_war_count == 0
    assert did.rarity_upgrades == 0
    assert did.is_pet
    assert did.reforge is None
    assert did.dungeon_stars == 0
    assert did.scrolls == []


def test_item_hyperion() -> None:
    raw_data = "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"
    did = DecodedNBT(raw_data)
    assert did
    assert did.skyblock_id == "HYPERION"
    assert did.enchantments == [
        "ENCHANTMENT_IMPALING_3",
        "ENCHANTMENT_LUCK_6",
        "ENCHANTMENT_CRITICAL_6",
        "ENCHANTMENT_CLEAVE_6",
        "ENCHANTMENT_SMOLDERING_2",
        "ENCHANTMENT_LOOTING_4",
        "ENCHANTMENT_SYPHON_4",
        "ENCHANTMENT_SMITE_7",
        "ENCHANTMENT_SCAVENGER_4",
        "ENCHANTMENT_ENDER_SLAYER_7",
        "ENCHANTMENT_FIRE_ASPECT_3",
        "ENCHANTMENT_VAMPIRISM_6",
        "ENCHANTMENT_EXPERIENCE_4",
        "ENCHANTMENT_EXECUTE_5",
        "ENCHANTMENT_GIANT_KILLER_6",
        "ENCHANTMENT_VENOMOUS_5",
        "ENCHANTMENT_DRAGON_HUNTER_1",
        "ENCHANTMENT_FIRST_STRIKE_4",
        "ENCHANTMENT_THUNDERLORD_7",
        "ENCHANTMENT_ULTIMATE_WISE_5",
        "ENCHANTMENT_CUBISM_5",
        "ENCHANTMENT_CHAMPION_10",
        "ENCHANTMENT_LETHALITY_6",
    ]
    assert did.art_of_war_count == 1
    assert did.rarity_upgrades == 1
    assert not did.is_pet
    assert did.reforge == "heroic"
    assert did.dungeon_stars == 9
    assert did.scrolls == [
        "WITHER_SHIELD_SCROLL",
        "SHADOW_WARP_SCROLL",
        "IMPLOSION_SCROLL",
    ]

    assert did.dict() == {
        "Count": 1,
        "Damage": 0,
        "id": 267,
        "tag": {
            "ExtraAttributes": {
                "ability_scroll": None,
                "art_of_war_count": 1,
                "champion_combat_xp": 72582071.63017021,
                "enchantments": {
                    "champion": 10,
                    "cleave": 6,
                    "critical": 6,
                    "cubism": 5,
                    "dragon_hunter": 1,
                    "ender_slayer": 7,
                    "execute": 5,
                    "experience": 4,
                    "fire_aspect": 3,
                    "first_strike": 4,
                    "giant_killer": 6,
                    "impaling": 3,
                    "lethality": 6,
                    "looting": 4,
                    "luck": 6,
                    "scavenger": 4,
                    "smite": 7,
                    "smoldering": 2,
                    "syphon": 4,
                    "thunderlord": 7,
                    "ultimate_wise": 5,
                    "vampirism": 6,
                    "venomous": 5,
                },
                "gems": {
                    "COMBAT_0": "FINE",
                    "COMBAT_0_gem": "SAPPHIRE",
                    "DEFENSIVE_0": {"quality": "PERFECT", "uuid": "84c26564-2981-431f-9051-273668a0af50"},
                    "DEFENSIVE_0_gem": "AMETHYST",
                    "SAPPHIRE_0": "FINE",
                    "unlocked_slots": None,
                },
                "hot_potato_count": 15,
                "id": "HYPERION",
                "modifier": "heroic",
                "rarity_upgrades": 1,
                "stats_book": 177662,
                "timestamp": 1657706160000,
                "upgrade_level": 9,
                "uuid": "21616c2d-6a14-4690-9b52-8d1857ed30cb",
            },
            "HideFlags": 254,
            "Unbreakable": 1,
            "display": {"Lore": None, "Name": "§dHeroic Hyperion §6✪§6✪§6✪§6✪§6✪§c➍"},
            "ench": None,
        },
    }


@pytest.mark.parametrize("raw_data", _BULK_NBTS)
def test_bulk_nbts(raw_data: str) -> None:
    DecodedNBT(raw_data)


@pytest.mark.parametrize("raw_data", _BULK_NBTS)
def test_bulk_nbts_dicts(raw_data: str) -> None:
    DecodedNBT(raw_data).dict()



import pytest

from reskyblock import Client


@pytest.fixture
def client() -> Client:
    return Client()


@pytest.mark.asyncio
async def test_client() -> None:
    _ = Client()


@pytest.mark.asyncio
async def test_get_auctions(client: Client) -> None:
    auctions = await client.get_auctions()
    assert auctions.success
    assert auctions.auctions


@pytest.mark.asyncio
async def test_get_auctions_ended(client: Client) -> None:
    ended_auctions = await client.get_auctions_ended()
    assert ended_auctions.success
    assert ended_auctions.auctions


@pytest.mark.asyncio
async def test_get_bazaar(client: Client) -> None:
    bazaar = await client.get_bazaar()
    assert bazaar.success
    assert bazaar.products


@pytest.mark.asyncio
async def test_get_all_auctions(client: Client) -> None:
    auctions = await client.get_all_auctions(max_pages=1)
    assert len(auctions.auctions) == 2000
    assert auctions.received_at > 0



import json
import pathlib

import pytest
from pytest_httpx import HTTPXMock

from reskyblock import Client

_AUCTIONS_DATA = (pathlib.Path(__file__).resolve().parents[0] / "data" / "auctions.json").read_text(encoding="utf-8")
_AUCTIONS_DATA2 = (pathlib.Path(__file__).resolve().parents[0] / "data" / "auctions2.json").read_text(encoding="utf-8")

_AUCTIONS_ENDED_DATA = (pathlib.Path(__file__).resolve().parents[0] / "data" / "auctions_ended.json").read_text(
    encoding="utf-8"
)
_BAZAAR_DATA = (pathlib.Path(__file__).resolve().parents[0] / "data" / "bazaar.json").read_text(encoding="utf-8")


@pytest.fixture
def client() -> Client:
    return Client()


@pytest.mark.asyncio
async def test_client() -> None:
    _ = Client()


@pytest.mark.asyncio
async def test_get_auctions(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(json=json.loads(_AUCTIONS_DATA))
    await client.get_auctions()


@pytest.mark.asyncio
async def test_get_auctions_ended(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(json=json.loads(_AUCTIONS_ENDED_DATA))

    await client.get_auctions_ended()


@pytest.mark.asyncio
async def test_get_bazaar(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(json=json.loads(_BAZAAR_DATA))

    await client.get_bazaar()


@pytest.mark.asyncio
async def test_continuous_update_auctions(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(json=json.loads(_AUCTIONS_DATA))

    async for auction in await client.get_auctions_continuous():
        assert len(auction.auctions) == 1000
        assert auction.received_at > 0
        break


@pytest.mark.asyncio
async def test_continuous_update_auctions_ended(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(json=json.loads(_AUCTIONS_ENDED_DATA))

    async for _ in await client.get_auctions_ended_continuous():
        break


@pytest.mark.asyncio
async def test_continuous_update_bazaar(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(json=json.loads(_BAZAAR_DATA))

    async for _ in await client.get_bazaar_continuous():
        break


@pytest.mark.asyncio
async def test_get_all_auctions(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=0", json=json.loads(_AUCTIONS_DATA))
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=1", json=json.loads(_AUCTIONS_DATA2))
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=2", status_code=404)

    auctions = await client.get_all_auctions()
    assert len(auctions.auctions) == 2000
    assert auctions.received_at > 0


@pytest.mark.asyncio
async def test_get_all_auctions_continuous(httpx_mock: HTTPXMock, client: Client) -> None:
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=0", json=json.loads(_AUCTIONS_DATA))
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=0", json=json.loads(_AUCTIONS_DATA))
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=1", json=json.loads(_AUCTIONS_DATA2))
    httpx_mock.add_response(url="https://api.hypixel.net/v2/skyblock/auctions?page=2", status_code=404)
    async for _ in await client.get_all_auctions_continuous():
        break



from __future__ import annotations

import msgspec

from reskyblock.nbt import DecodedNBT

__all__ = (
    "Auction",
    "Auctions",
    "Bid",
)


class Bid(msgspec.Struct):
    auction_id: str
    bidder: str
    profile_id: str
    amount: int
    timestamp: int


class Auctions(msgspec.Struct, rename="camel"):
    success: bool
    page: int
    total_pages: int
    total_auctions: int
    last_updated: int
    auctions: list[Auction]
    received_at: float = 0


class Auction(msgspec.Struct):
    """Represents an auction instance on the auction house"""

    start: int
    end: int
    item_name: str
    extra: str
    category: str
    tier: str
    starting_bid: int
    item_bytes: str
    claimed: bool
    highest_bid_amount: int
    last_updated: int
    bin: bool
    uuid: str
    auctioneer: str
    profile_id: str
    decoded_nbt: DecodedNBT | None = None
    command: str | None = None

    def __post_init__(self) -> None:
        if self.decoded_nbt is None:
            self.decoded_nbt = DecodedNBT(raw_data=self.item_bytes)

        if self.command is None:
            self.command = (
                f"/viewauction {self.uuid[:9]}-"
                f"{self.uuid[9:13]}-{self.uuid[13:17]}-"
                f"{self.uuid[17:21]}-{self.uuid[21:]}"
            )


class AllAuctions(msgspec.Struct):
    last_updated: int
    auctions: list[Auction]
    received_at: float = 0



import pytest
from pytest_httpx import HTTPXMock

from reskyblock.http import HTTPXAsyncClient, HTTPXClient


@pytest.mark.parametrize("params", [None, {"a": "a", "b": "b"}])
def test_httpx_client(httpx_mock: HTTPXMock, params: dict[str, str] | None) -> None:
    httpx_mock.add_response()

    data = HTTPXClient().get("https://test_url", params=params)
    assert data == b""


@pytest.mark.asyncio
@pytest.mark.parametrize("params", [None, {"a": "a", "b": "b"}])
async def test_async_httpx_client(httpx_mock: HTTPXMock, params: dict[str, str] | None) -> None:
    httpx_mock.add_response()

    data = await HTTPXAsyncClient().get("https://test_url", params=params)
    assert data == b""



"""URLs for the SkyBlock API"""


_AUCTIONS_URL = "https://api.hypixel.net/v2/skyblock/auctions"
_AUCTIONS_ENDED_URL = "https://api.hypixel.net/v2/skyblock/auctions_ended"
_BAZAAR_URL = "https://api.hypixel.net/v2/skyblock/bazaar"


def _prepare_auctions_url(page: int | None = None) -> str:
    if page is None:
        page = 0
    return f"{_AUCTIONS_URL}?page={page}"


def _prepare_auctions_ended_url() -> str:
    return _AUCTIONS_ENDED_URL


def _prepare_bazaar_url() -> str:
    return _BAZAAR_URL



import msgspec

from reskyblock.nbt import DecodedNBT

__all__ = (
    "AuctionsEnded",
    "EndedAuction",
)


class EndedAuction(msgspec.Struct):
    """Represent an instance of an auction that has ended"""

    auction_id: str
    seller: str
    seller_profile: str
    buyer: str
    timestamp: int
    price: int
    bin: bool
    item_bytes: str
    decoded_nbt: DecodedNBT | None = None

    def __post_init__(self) -> None:
        if self.decoded_nbt is None:
            self.decoded_nbt = DecodedNBT(raw_data=self.item_bytes)


class AuctionsEnded(msgspec.Struct, rename="camel"):
    success: bool
    last_updated: int
    auctions: list[EndedAuction]
    received_at: float = 0



import asyncio
import logging
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Awaitable, Callable
from functools import partial

from httpx import HTTPStatusError

from reskyblock.http import AbstractAsyncHTTPClient, HTTPXAsyncClient
from reskyblock.models import (
    AllAuctions,
    Auction,
    Auctions,
    AuctionsEnded,
    Bazaar,
    EndedAuction,
    Product,
    QuickStatus,
    Summary,
)
from reskyblock.nbt import DecodedNBT
from reskyblock.serialization import AbstractJSONDecoder, MSGSpecDecoder
from reskyblock.urls import _prepare_auctions_ended_url, _prepare_auctions_url, _prepare_bazaar_url

type APIEndpoint = Auctions | AuctionsEnded | Bazaar | AllAuctions
type APIEndpointGetter = Callable[[], Awaitable[APIEndpoint]]

__all__ = ("Client", "MockClient", "AbstractClient")


class AbstractClient(ABC):
    @abstractmethod
    async def get_auctions(self, page: int = 0) -> Auctions:
        pass

    @abstractmethod
    async def get_auctions_ended(self) -> AuctionsEnded:
        pass

    @abstractmethod
    async def get_bazaar(self) -> Bazaar:
        pass

    @abstractmethod
    async def get_all_auctions(self, max_pages: int = 100) -> AllAuctions:
        pass

    @abstractmethod
    async def get_auctions_continuous(self) -> AsyncIterator[Auctions]:
        pass

    @abstractmethod
    async def get_auctions_ended_continuous(self) -> AsyncIterator[AuctionsEnded]:
        pass

    @abstractmethod
    async def get_bazaar_continuous(self) -> AsyncIterator[Bazaar]:
        pass

    @abstractmethod
    async def get_all_auctions_continuous(self, max_pages: int = 100) -> AsyncIterator[AllAuctions]:
        pass


class Client(AbstractClient):
    def __init__(self) -> None:
        self._http_client: AbstractAsyncHTTPClient = HTTPXAsyncClient()
        self._json_decoder: AbstractJSONDecoder = MSGSpecDecoder()
        self._auctions_last_updated: int = 0
        self._auctions_ended_last_updated: int = 0
        self._bazaar_last_updated: int = 0

    async def get_auctions(self, page: int = 0) -> Auctions:
        """Get a single page of active auctions"""
        resp_bytes = await self._http_client.get(url=_prepare_auctions_url(page))
        auctions: Auctions = self._json_decoder.serialize(resp_bytes, Auctions)
        auctions.received_at = time.time() * 1000
        self._auctions_last_updated = auctions.last_updated
        return auctions

    async def get_auctions_ended(self) -> AuctionsEnded:
        """Get ended auctions"""
        resp_bytes = await self._http_client.get(url=_prepare_auctions_ended_url())
        auctions_ended: AuctionsEnded = self._json_decoder.serialize(resp_bytes, AuctionsEnded)
        auctions_ended.received_at = time.time() * 1000
        self._auctions_ended_last_updated = auctions_ended.last_updated
        return auctions_ended

    async def get_bazaar(self) -> Bazaar:
        """Get bazaar endpoint"""
        resp_bytes = await self._http_client.get(url=_prepare_bazaar_url())
        bazaar: Bazaar = self._json_decoder.serialize(resp_bytes, Bazaar)
        bazaar.received_at = time.time() * 1000
        self._bazaar_last_updated = bazaar.last_updated
        return bazaar

    async def get_all_auctions(self, max_pages: int = 100) -> AllAuctions:
        """Get auctions from all pages"""
        auctions = []
        page = 0
        last_updated = 0
        while page <= max_pages:
            try:
                auctions_page = await self.get_auctions(page)
                auctions.extend(auctions_page.auctions)
                last_updated = auctions_page.last_updated
                page += 1
            except HTTPStatusError:
                break
        return AllAuctions(last_updated, auctions, time.time() * 1000)

    @staticmethod
    async def _get_continuous[T: APIEndpoint](
        getter: APIEndpointGetter, expected_update_interval: float, update_getter: APIEndpointGetter | None = None
    ) -> AsyncIterator[T]:
        use_update_getter_for_return = update_getter is None
        if update_getter is None:
            update_getter = getter

        last_updated = 0
        while 1:
            next_update = last_updated / 1000 + expected_update_interval
            if next_update > time.time():  # the next update is in the future
                sleep_for = next_update - time.time()
                await asyncio.sleep(max(sleep_for, 0.1))
            try:
                update_api_endpoint = await update_getter()
            except Exception as e:
                logging.exception(e)
                continue

            if update_api_endpoint.last_updated == last_updated:
                continue  # the API has not updated yet

            last_updated = update_api_endpoint.last_updated
            if use_update_getter_for_return:
                api_endpoint = update_api_endpoint
            else:
                try:
                    api_endpoint = await getter()
                except Exception as e:
                    logging.exception(e)
                    continue
            yield api_endpoint

    async def get_auctions_continuous(self) -> AsyncIterator[Auctions]:
        return self._get_continuous(self.get_auctions, 66.5)

    async def get_auctions_ended_continuous(self) -> AsyncIterator[AuctionsEnded]:
        return self._get_continuous(self.get_auctions_ended, 60)

    async def get_bazaar_continuous(self) -> AsyncIterator[Bazaar]:
        return self._get_continuous(self.get_bazaar, 20)

    async def get_all_auctions_continuous(self, max_pages: int = 100) -> AsyncIterator[AllAuctions]:
        return self._get_continuous(partial(self.get_all_auctions, max_pages), 66.5, self.get_auctions)


class MockClient(AbstractClient):
    """This is mock client intended for the users of the library to test code that relies on
    the reskyblock client.

    The getters return a minimal complete representation of each API page. The continuous producers
    yield only a single item.
    """

    async def get_auctions(self, page: int = 0) -> Auctions:
        return Auctions(
            success=True,
            page=0,
            total_pages=1,
            total_auctions=1,
            last_updated=0,
            auctions=[
                Auction(
                    start=0,
                    end=0,
                    item_name="mock item",
                    extra="",
                    category="",
                    tier="",
                    starting_bid=1,
                    item_bytes="",
                    claimed=False,
                    highest_bid_amount=0,
                    last_updated=0,
                    bin=True,
                    uuid="",
                    auctioneer="",
                    profile_id="",
                    decoded_nbt=DecodedNBT(
                        raw_data="",
                        skyblock_id="MOCK_ITEM",
                    ),
                    command="",
                )
            ],
        )

    async def get_auctions_ended(self) -> AuctionsEnded:
        return AuctionsEnded(
            success=True,
            last_updated=0,
            auctions=[
                EndedAuction(
                    auction_id="",
                    seller="",
                    seller_profile="",
                    buyer="",
                    timestamp=0,
                    price=1,
                    bin=True,
                    item_bytes="",
                    decoded_nbt=DecodedNBT(
                        raw_data="",
                        skyblock_id="MOCK_ITEM",
                    ),
                )
            ],
        )

    async def get_bazaar(self) -> Bazaar:
        return Bazaar(
            success=True,
            last_updated=0,
            products={
                "MOCK_PRODUCT": Product(
                    product_id="MOCK_PRODUCT",
                    sell_summary=[
                        Summary(
                            amount=1,
                            price_per_unit=1.0,
                            orders=1,
                        )
                    ],
                    buy_summary=[
                        Summary(
                            amount=1,
                            price_per_unit=1.0,
                            orders=1,
                        )
                    ],
                    quick_status=QuickStatus(
                        product_id="MOCK_PRODUCT",
                        sell_price=1.0,
                        sell_volume=0,
                        sell_moving_week=0,
                        sell_orders=0,
                        buy_price=1.0,
                        buy_volume=0,
                        buy_moving_week=0,
                        buy_orders=0,
                    ),
                )
            },
        )

    async def get_all_auctions(self, max_pages: int = 100) -> AllAuctions:
        auctions = await self.get_auctions()
        return AllAuctions(last_updated=0, auctions=auctions.auctions)

    async def get_auctions_continuous(self) -> AsyncIterator[Auctions]:
        return self._iterator_single_item(self.get_auctions)

    async def get_auctions_ended_continuous(self) -> AsyncIterator[AuctionsEnded]:
        return self._iterator_single_item(self.get_auctions_ended)

    async def get_bazaar_continuous(self) -> AsyncIterator[Bazaar]:
        return self._iterator_single_item(self.get_bazaar)

    async def get_all_auctions_continuous(self, max_pages: int = 100) -> AsyncIterator[AllAuctions]:
        return self._iterator_single_item(self.get_all_auctions)

    @staticmethod
    async def _iterator_single_item[T: APIEndpoint](func: APIEndpointGetter) -> AsyncIterator[T]:
        """This is a helper function that creates an async iterator, which produces only one item.
        The produced item is the result of an `APIEndpoint` getter function.
        """
        yield await func()



"""This module provides all needed functionality for dealing with http requests"""

from abc import ABC, abstractmethod

import httpx

__all__ = ("AbstractHTTPClient", "HTTPXClient", "HTTPXAsyncClient", "AbstractAsyncHTTPClient")


class AbstractHTTPClient(ABC):
    """Abstract interface for an HTTP client"""

    @abstractmethod
    def get(self, url: str, params: dict | None = None) -> bytes:
        """Sends a GET HTTP request to a URL
        :param url: URL to be fetched
        :param params: parameters to be added to the request
        :return: response as bytes
        """


class AbstractAsyncHTTPClient(ABC):
    """Abstract interface for an asynchronous HTTP client"""

    @abstractmethod
    async def get(self, url: str, params: dict | None = None) -> bytes:
        """Sends a GET HTTP request to a URL
        :param url: URL to be fetched
        :param params: parameters to be added to the request
        :return: response as bytes
        """


class HTTPXClient(AbstractHTTPClient):
    def __init__(self) -> None:
        self._session: httpx.Client = httpx.Client()

    def get(self, url: str, params: dict | None = None) -> bytes:
        """Sends a GET HTTP request to a URL. If an exception is thrown,
        retries sending the same request after `retry_time` seconds.
        :param url: URL to be fetched
        :param params: parameters to be added to the request
        :return: response as bytes
        """

        return self._session.get(url, params=params).raise_for_status().content


class HTTPXAsyncClient(AbstractAsyncHTTPClient):
    def __init__(self) -> None:
        self._session: httpx.AsyncClient = httpx.AsyncClient()

    async def get(self, url: str, params: dict | None = None) -> bytes:
        """Sends a GET HTTP request to a URL. If an exception is thrown,
        retries sending the same request after `retry_time` seconds.
        :param url: URL to be fetched
        :param params: parameters to be added to the request
        :return: response as bytes
        """

        return (await self._session.get(url, params=params)).raise_for_status().content



import pytest

from reskyblock.urls import _prepare_auctions_ended_url, _prepare_auctions_url, _prepare_bazaar_url


def test__prepare_auctions_ended_url() -> None:
    assert _prepare_auctions_ended_url() == "https://api.hypixel.net/v2/skyblock/auctions_ended"


@pytest.mark.parametrize("page", [0, 1, 100])
def test__prepare_auctions_url(page: int | None) -> None:
    assert _prepare_auctions_url(page) == f"https://api.hypixel.net/v2/skyblock/auctions?page={page}"


def test__prepare_auctions_url_none() -> None:
    assert _prepare_auctions_url() == "https://api.hypixel.net/v2/skyblock/auctions?page=0"


def test__prepare_bazaar_url() -> None:
    assert _prepare_bazaar_url() == "https://api.hypixel.net/v2/skyblock/bazaar"
